@@ -67,6 +67,7 @@ typedef struct _GKeyFileGroup GKeyFileGroup;
 struct _GKeyFile
 {
   GList *groups;
+  GHashTable *group_hash;
 
   GKeyFileGroup *start_group;
   GKeyFileGroup *current_group;
@@ -80,6 +81,8 @@ struct _GKeyFile
   gchar list_separator;
 
   GKeyFileFlags flags;
+
+  gchar **locales;
 };
 
 typedef struct _GKeyFileKeyValuePair GKeyFileKeyValuePair;
@@ -199,11 +202,13 @@ g_key_file_init (GKeyFile *key_file)
 {  
   key_file->current_group = g_slice_new0 (GKeyFileGroup);
   key_file->groups = g_list_prepend (NULL, key_file->current_group);
+  key_file->group_hash = g_hash_table_new (g_str_hash, g_str_equal);
   key_file->start_group = NULL;
   key_file->parse_buffer = g_string_sized_new (128);
   key_file->approximate_size = 0;
   key_file->list_separator = ';';
   key_file->flags = 0;
+  key_file->locales = g_strdupv ((gchar **)g_get_language_names ());
 }
 
 static void
@@ -211,8 +216,17 @@ g_key_file_clear (GKeyFile *key_file)
 {
   GList *tmp, *group_node;
 
+  if (key_file->locales) 
+    {
+      g_strfreev (key_file->locales);
+      key_file->locales = NULL;
+    }
+
   if (key_file->parse_buffer)
-    g_string_free (key_file->parse_buffer, TRUE);
+    {
+      g_string_free (key_file->parse_buffer, TRUE);
+      key_file->parse_buffer = NULL;
+    }
 
   tmp = key_file->groups;
   while (tmp != NULL)
@@ -222,6 +236,9 @@ g_key_file_clear (GKeyFile *key_file)
       g_key_file_remove_group_node (key_file, group_node);
     }
 
+  g_hash_table_destroy (key_file->group_hash);
+  key_file->group_hash = NULL;
+
   g_assert (key_file->groups == NULL);
 }
 
@@ -540,7 +557,7 @@ g_key_file_load_from_data (GKeyFile       *key_file,
 }
 
 /**
- * g_key_file_load_from_data_dirs:
+ * g_key_file_load_from_dirs:
  * @key_file: an empty #GKeyFile struct
  * @file: a relative path to a filename to open and parse
  * @search_dirs: %NULL-terminated array of directories to search
@@ -695,17 +712,14 @@ static gboolean
 g_key_file_locale_is_interesting (GKeyFile    *key_file,
 				  const gchar *locale)
 {
-  const gchar * const * current_locales;
   gsize i;
 
   if (key_file->flags & G_KEY_FILE_KEEP_TRANSLATIONS)
     return TRUE;
 
-  current_locales = g_get_language_names ();
-
-  for (i = 0; current_locales[i] != NULL; i++)
+  for (i = 0; key_file->locales[i] != NULL; i++)
     {
-      if (g_ascii_strcasecmp (current_locales[i], locale) == 0)
+      if (g_ascii_strcasecmp (key_file->locales[i], locale) == 0)
 	return TRUE;
     }
 
@@ -1461,6 +1475,9 @@ g_key_file_get_string_list (GKeyFile     *key_file,
   g_return_val_if_fail (group_name != NULL, NULL);
   g_return_val_if_fail (key != NULL, NULL);
 
+  if (length)
+    *length = 0;
+
   value = g_key_file_get_value (key_file, group_name, key, &key_file_error);
 
   if (key_file_error)
@@ -1743,7 +1760,11 @@ g_key_file_get_locale_string_list (GKeyFile     *key_file,
     g_propagate_error (error, key_file_error);
   
   if (!value)
-    return NULL;
+    {
+      if (length)
+        *length = 0;
+      return NULL;
+    }
 
   if (value[strlen (value) - 1] == ';')
     value[strlen (value) - 1] = '\0';
@@ -1939,6 +1960,9 @@ g_key_file_get_boolean_list (GKeyFile     *key_file,
   g_return_val_if_fail (group_name != NULL, NULL);
   g_return_val_if_fail (key != NULL, NULL);
 
+  if (length)
+    *length = 0;
+
   key_file_error = NULL;
 
   values = g_key_file_get_string_list (key_file, group_name, key,
@@ -2151,6 +2175,9 @@ g_key_file_get_integer_list (GKeyFile     *key_file,
   g_return_val_if_fail (group_name != NULL, NULL);
   g_return_val_if_fail (key != NULL, NULL);
 
+  if (length)
+    *length = 0;
+
   values = g_key_file_get_string_list (key_file, group_name, key,
 				       &num_ints, &key_file_error);
 
@@ -2360,6 +2387,9 @@ g_key_file_get_double_list  (GKeyFile     *key_file,
   g_return_val_if_fail (group_name != NULL, NULL);
   g_return_val_if_fail (key != NULL, NULL);
 
+  if (length)
+    *length = 0;
+
   values = g_key_file_get_string_list (key_file, group_name, key,
                                        &num_doubles, &key_file_error);
 
@@ -2767,8 +2797,8 @@ g_key_file_get_group_comment (GKeyFile             *key_file,
   GList *group_node;
   GKeyFileGroup *group;
   
-  group_node = g_key_file_lookup_group_node (key_file, group_name);
-  if (!group_node)
+  group = g_key_file_lookup_group (key_file, group_name);
+  if (!group)
     {
       g_set_error (error, G_KEY_FILE_ERROR,
                    G_KEY_FILE_ERROR_GROUP_NOT_FOUND,
@@ -2778,10 +2808,10 @@ g_key_file_get_group_comment (GKeyFile             *key_file,
       return NULL;
     }
 
-  group = (GKeyFileGroup *)group_node->data;
   if (group->comment)
     return g_strdup (group->comment->value);
   
+  group_node = g_key_file_lookup_group_node (key_file, group_name);
   group_node = group_node->next;
   group = (GKeyFileGroup *)group_node->data;  
   return get_group_comment (key_file, group, error);
@@ -2888,7 +2918,7 @@ g_key_file_has_group (GKeyFile    *key_file,
   g_return_val_if_fail (key_file != NULL, FALSE);
   g_return_val_if_fail (group_name != NULL, FALSE);
 
-  return g_key_file_lookup_group_node (key_file, group_name) != NULL;
+  return g_key_file_lookup_group (key_file, group_name) != NULL;
 }
 
 /**
@@ -2961,6 +2991,8 @@ g_key_file_add_group (GKeyFile    *key_file,
 
   if (key_file->start_group == NULL)
     key_file->start_group = group;
+
+  g_hash_table_insert (key_file->group_hash, (gpointer)group->name, group);
 }
 
 static void
@@ -3015,6 +3047,9 @@ g_key_file_remove_group_node (GKeyFile *key_file,
 
   group = (GKeyFileGroup *) group_node->data;
 
+  if (group->name)
+    g_hash_table_remove (key_file->group_hash, group->name);
+
   /* If the current group gets deleted make the current group the last
    * added group.
    */
@@ -3207,14 +3242,7 @@ static GKeyFileGroup *
 g_key_file_lookup_group (GKeyFile    *key_file,
 			 const gchar *group_name)
 {
-  GList *group_node;
-
-  group_node = g_key_file_lookup_group_node (key_file, group_name);
-
-  if (group_node != NULL)
-    return (GKeyFileGroup *) group_node->data; 
-
-  return NULL;
+  return (GKeyFileGroup *)g_hash_table_lookup (key_file->group_hash, group_name);
 }
 
 static GList *
@@ -3267,7 +3295,7 @@ g_key_file_is_group_name (const gchar *name)
 
   p = q = (gchar *) name;
   while (*q && *q != ']' && *q != '[' && !g_ascii_iscntrl (*q))
-    q = g_utf8_next_char (q);
+    q = g_utf8_find_next_char (q, NULL);
   
   if (*q != '\0' || q == p)
     return FALSE;
@@ -3288,14 +3316,14 @@ g_key_file_is_key_name (const gchar *name)
    * since gnome-vfs uses mime-types as keys in its cache.
    */
   while (*q && *q != '=' && *q != '[' && *q != ']')
-    q = g_utf8_next_char (q);
+    q = g_utf8_find_next_char (q, NULL);
   
   /* No empty keys, please */
   if (q == p)
     return FALSE;
 
   /* We accept spaces in the middle of keys to not break
-   * existing apps, but we don't tolerate initial of final
+   * existing apps, but we don't tolerate initial or final
    * spaces, which would lead to silent corruption when
    * rereading the file.
    */
@@ -3305,8 +3333,8 @@ g_key_file_is_key_name (const gchar *name)
   if (*q == '[')
     {
       q++;
-      while (*q && (g_unichar_isalnum (g_utf8_get_char (q)) || *q == '-' || *q == '_' || *q == '.' || *q == '@'))
-        q = g_utf8_next_char (q);
+      while (*q && (g_unichar_isalnum (g_utf8_get_char_validated (q, -1)) || *q == '-' || *q == '_' || *q == '.' || *q == '@'))
+        q = g_utf8_find_next_char (q, NULL);
 
       if (*q != ']')
         return FALSE;     
@@ -3335,15 +3363,15 @@ g_key_file_line_is_group (const gchar *line)
   p++;
 
   while (*p && *p != ']')
-    p = g_utf8_next_char (p);
+    p = g_utf8_find_next_char (p, NULL);
 
   if (*p != ']')
     return FALSE;
  
   /* silently accept whitespace after the ] */
-  p = g_utf8_next_char (p);
+  p = g_utf8_find_next_char (p, NULL);
   while (*p == ' ' || *p == '\t')
-    p = g_utf8_next_char (p);
+    p = g_utf8_find_next_char (p, NULL);
      
   if (*p)
     return FALSE;
@@ -3565,7 +3593,7 @@ g_key_file_parse_value_as_integer (GKeyFile     *key_file,
 				   GError      **error)
 {
   gchar *end_of_valid_int;
-  glong long_value;
+ glong long_value;
   gint int_value;
 
   errno = 0;
